import java.util.*;
import java.io.*;

public class ExactInference extends Inference {
    
    public ExactInference(Node[] ns,int doDebug) {
	super(ns,doDebug);
    }//constructor

    /***********     Variable Elimination WITHOUT evidence     **********/
    //the parameter query is an UNINSTATIATED expression
    public void variableElimination(String query) {

        //check for prior probabilities of nodes with no parents
        //these probabilities were provided to us when the bayes net was created
        double test = findProbInGlobalCPT(query+"=true");
        if (test != 1.0) {
            System.out.println("Base Case: P("+query+"=true)="+ test);
            return;
        }
        
	Vector _F = initialiseF("");
	Vector _Z = initialiseZ(query);
        if(DEBUG==1) {
            System.out.println("F="+_F);
            System.out.println("Z="+_Z);
        }
        
        outerfor: for (int i=0; i<_Z.size(); i++) {
            //choose a node to eliminate
	    String _eliminateThisNode = (String)_Z.get(i);
	    if (DEBUG==1) System.out.println("\n*>eliminateThisNode="+_eliminateThisNode);

            //note that the method initialiseFprime modifies F
	    Vector _Fprime = initialiseFprime(_F,_eliminateThisNode);

            
            /****************************  CALCULATE NEW FACTORS  ******************************/
	    /**** The new factors that are calculated will include all the variables        ****/
	    /**** that are currently uninstantiated. eliminateThisVariable cannot be a part ****/
	    /**** of any new factor since it is being eliminated, so we disinclude it.      ****/
	    Vector _U = countUninstantiated(_Fprime,_eliminateThisNode);
	    int _numU = _U.size();
	    if (DEBUG==1) System.out.println("Uninstantiated Variables="+_U);
	    /***********************************************************************************/

	    /*******************************  INITIALISE F''  **********************************/
	    /****  partially instantiate F' by assigning the variable we want to eliminate  ****/
	    /****  it's values (ie, true or false). For example _Fprimeprime_TRUE will      ****/
	    /****  contain entries of the form wetgrass=true|rain=false if the variable     ****/
	    /****  we are currently eliminating is wetgrass and rain is a query variable    ****/
	    /***** We split F'' into F''TRUE and F''FALSE such that F''=F''TRUE + F''FALSE  ****/
	    Vector _Fprimeprime_TRUE = instantiateVariable(_eliminateThisNode,_TRUE,_Fprime);
	    Vector _Fprimeprime_FALSE = instantiateVariable(_eliminateThisNode,_FALSE,_Fprime);
            if (DEBUG==1) {
                System.out.println("F''_TRUE="+_Fprimeprime_TRUE);
                System.out.println("F''_FALSE="+_Fprimeprime_FALSE);
            }
	    /***********************************************************************************/

            //the number of new factors is a power of 2 of the number of uninst variables 
            innerfor: for (int p=0; p<Math.pow(2,_numU); p++) {
                String newFactor_instantiated =  "";
                String newFactor_uninstantiated = "";
                
                int[] binary = int2binary(p,_numU); //generate one instantiation
                Vector _copyOfEIT = (Vector)_Fprimeprime_TRUE.clone();
                Vector _copyOfEIF = (Vector)_Fprimeprime_FALSE.clone();
                Enumeration _e = _U.elements();
                
                innerinnerfor: for (int r=0; r<_numU; r++) {
                    String currentVariable = (String)_e.nextElement();
                    
                    if (binary[r] == 0) {
                        _copyOfEIF = instantiateVariable(currentVariable,_FALSE,_copyOfEIF);
                        _copyOfEIT = instantiateVariable(currentVariable,_FALSE,_copyOfEIT);
                        newFactor_instantiated = newFactor_instantiated.concat(currentVariable+_FALSE);
                    }
                    else {
                        _copyOfEIT = instantiateVariable(currentVariable,_TRUE,_copyOfEIT);
                        _copyOfEIF = instantiateVariable(currentVariable,_TRUE,_copyOfEIF);
                        newFactor_instantiated = newFactor_instantiated.concat(currentVariable+_TRUE);
                        
                    }
                    newFactor_uninstantiated = newFactor_uninstantiated.concat(currentVariable);
                    if (r+1 < _numU) {
                        newFactor_instantiated = newFactor_instantiated.concat("&");
                        newFactor_uninstantiated = newFactor_uninstantiated.concat("&");
                    }
                }//innerinner for
                
                double D_EIF = multiplyVector(_copyOfEIF);
                double D_EIT  = multiplyVector(_copyOfEIT);
                double probability = D_EIF+D_EIT;
                
                addFactorToGlobalCPT(newFactor_instantiated,String.valueOf(probability));
                
                if (!contains(_F,newFactor_uninstantiated))
                    _F.add(newFactor_uninstantiated);
                
                if (DEBUG==1) System.out.println("New Factor("+newFactor_instantiated+")="+probability);
            }//inner for
	}//outer for
        
	if (DEBUG==1) System.out.println("\noutside for loops: F="+_F);

	//the factors that remain in F have to be multiplied together
        normaliseForNoEvidence(_F);

	reinitialiseGlobalCPT();

    }//variableElimination


    /***********       Variable Elimination with evidence       **********/
    //the parameter query is UN-instantiated and evidence is instantiated
    public void variableEliminationEvidence(String query,String evidence) {

        //check for conditional probabilities that were provided to us
        //when the bayes net was created
        double test = findProbInGlobalCPT(query+"=true|"+evidence);
        if (test != 1.0) {
            System.out.println("NO Variable Elimination: P("+query+"=true|"+evidence+")="+ test );
            return;
        }
		  test = findProbInGlobalCPT(evidence+"|"+query+"=true");
		  if (test != 1.0) {
		  		test *= findProbInGlobalCPT(query+"=true");
				test = test / findProbInGlobalCPT(evidence);
				System.out.println("No Variable Elimination: P("+query+"=true|"+evidence+")="+test);
				return;
		  }
        
	Vector _F = initialiseF(evidence);
	Vector _Z = initialiseZ(query+evidence);
        
        if(DEBUG==1) {
            System.out.println("F="+_F);
            System.out.println("Z="+_Z);
        }
        

        outerfor: for (int i=0; i<_Z.size(); i++) {
            
	    //choose a node to eliminate
	    String _eliminateThisNode = (String)_Z.get(i);
	    if (DEBUG==1) System.out.println("\n*>eliminateThisNode="+_eliminateThisNode);

            //note that the method initialiseFprime modifies F
	    Vector _Fprime = initialiseFprime(_F,_eliminateThisNode);

            
	    /****************************  CALCULATE NEW FACTORS  ******************************/
	    /**** The new factors that are calculated will include all the variables        ****/
	    /**** that are currently uninstantiated. eliminateThisVariable cannot be a part ****/
	    /**** of any new factor since it is being eliminated, so we disinclude it.      ****/
	    Vector _U = countUninstantiated(_Fprime,_eliminateThisNode);
	    int _numU = _U.size();
	    if (DEBUG==1) System.out.println("Uninstantiated Variables="+_U);
	    /***********************************************************************************/

            
	    /*******************************  INITIALISE F''  **********************************/
	    /****  partially instantiate F' by assigning the variable we want to eliminate  ****/
	    /****  it's values (ie, true or false). For example _Fprimeprime_TRUE will      ****/
	    /****  contain entries of the form wetgrass=true|rain=false if the variable     ****/
	    /****  we are currently eliminating is wetgrass and rain is a query variable    ****/
	    /***** We split F'' into F''TRUE and F''FALSE such that F''=F''TRUE + F''FALSE  ****/
	    Vector _Fprimeprime_TRUE = instantiateVariable(_eliminateThisNode,_TRUE,_Fprime);
	    Vector _Fprimeprime_FALSE = instantiateVariable(_eliminateThisNode,_FALSE,_Fprime);
            if (DEBUG==1) {
                System.out.println("F''_TRUE="+_Fprimeprime_TRUE);
                System.out.println("F''_FALSE="+_Fprimeprime_FALSE);
            }
	    /***********************************************************************************/

            //the number of new factors is a power of 2 of the number of uninst variables
            innerfor: for (int p=0; p<Math.pow(2,_numU); p++) {
                
                String newFactor_instantiated = initialiseFactor(evidence,_Fprime);
                String newFactor_uninstantiated = initialiseFactor(evidence,_Fprime);
                //System.out.println("instantiated string="+newFactor_instantiated);
                //System.out.println("uninstantiated string="+newFactor_uninstantiated);
                
                int[] binary = int2binary(p,_numU); //generate one instantiation
                Vector _copyOfEIT = (Vector)_Fprimeprime_TRUE.clone();
                Vector _copyOfEIF = (Vector)_Fprimeprime_FALSE.clone();
                Enumeration _e = _U.elements();
                
                innerinnerfor: for (int r=0; r<_numU; r++) {
                    String currentVariable = (String)_e.nextElement();
                    
                    if (binary[r] == 0) {
                        _copyOfEIF = instantiateVariable(currentVariable,_FALSE,_copyOfEIF);
                        _copyOfEIT = instantiateVariable(currentVariable,_FALSE,_copyOfEIT);
                        newFactor_instantiated = newFactor_instantiated.concat(currentVariable+_FALSE);
                    }
                    else {
                        _copyOfEIT = instantiateVariable(currentVariable,_TRUE,_copyOfEIT);
                        _copyOfEIF = instantiateVariable(currentVariable,_TRUE,_copyOfEIF);
                        newFactor_instantiated = newFactor_instantiated.concat(currentVariable+_TRUE);
                        
                    }
                    newFactor_uninstantiated = newFactor_uninstantiated.concat(currentVariable);
                    if (r+1 < _numU) {
                        newFactor_instantiated = newFactor_instantiated.concat("&");
                        newFactor_uninstantiated = newFactor_uninstantiated.concat("&");
                    }
                }//innerinner for
                
                double D_EIF = multiplyVector(_copyOfEIF);
                double D_EIT  = multiplyVector(_copyOfEIT);
                double probability = D_EIF+D_EIT;
                
                addFactorToGlobalCPT(newFactor_instantiated,String.valueOf(probability));
                
                if (!contains(_F,newFactor_uninstantiated))
                    _F.add(newFactor_uninstantiated);
                
                if (DEBUG==1) System.out.println("New Factor("+newFactor_instantiated+")="+probability);
            }//inner for
	}//outer for
        
	if (DEBUG==1) System.out.println("\noutside for loops: F="+_F);

        normaliseForEvidence(_F,evidence,query);
        
	reinitialiseGlobalCPT();

    }//variableEliminationEvidence


    /****************************************************************************/
    /***********                   HELPER METHODS                      **********/
    /****************************************************************************/

    /******************************initialiseFactor******************************/
    //the parameters fromThis and v are instantiated expressions
    //the string returned is the and-ed String Representation of the variables
    //that are common to both fromThis and v. If andAtEnd indicates whether an
    //and (ie, &) should be included right at the end; for example
    //rain=true&sprinkler=true&
    public String initialiseFactor(String fromThis, Vector v) 
    {
        String newFactor_instantiated = "";
        StringTokenizer ft = new StringTokenizer(fromThis,"&=|");
        while (ft.hasMoreTokens()) {
            String var = ft.nextToken();
            String varValue = ft.nextToken();
            if (vectorMentionsString(v,var))
                newFactor_instantiated =  newFactor_instantiated.concat(var+"="+varValue+"&");
            
        }//while
	if (newFactor_instantiated.equals("")) return newFactor_instantiated;
        return newFactor_instantiated;
    }//initialiseFactor
 
    /***************************normalise for NO evidence***********************/
    public void normaliseForNoEvidence(Vector _Fremain) 
    {
        Vector _U = countUninstantiated(_Fremain,"dummyString");
        int _numU = _U.size();
        if (DEBUG==1) System.out.println("Uninstantiated Variables="+_U);
        
        for (int p=0; p<Math.pow(2,_numU); p++) {

            String instRep = "";
            int[] binary = int2binary(p,_numU); //generate one instantiation
            Vector _copyOfF = (Vector)_Fremain.clone();
            Enumeration e = _U.elements();
            
            for (int r=0; r<_numU; r++) {
                String currentVariable = (String)e.nextElement();
                
                if (binary[r] == 0) {
                    _copyOfF = instantiateVariable(currentVariable,_FALSE,_copyOfF);
                    instRep = instRep.concat(currentVariable+_FALSE);
                }
                else {
                    _copyOfF = instantiateVariable(currentVariable,_TRUE,_copyOfF);
                    instRep = instRep.concat(currentVariable+_TRUE);
                    
                }
                if (r+1 < _numU) 
                    instRep = instRep.concat("&");
                
            }//printForNoEvidence
            double answer = multiplyVector(_copyOfF);
            System.out.println("P("+instRep+")="+answer);
        }//for
        
    }//normaliseForNoEvidence

    /***************************normalise for evidence**************************/
    public void normaliseForEvidence(Vector _Fremain, String evidence,String query) 
    {
        double totalsum = 0.0;
        
        Vector _U = countUninstantiated(_Fremain,"dummyString");
        int _numU = _U.size();
        if (DEBUG==1) System.out.println("Uninstantiated Variables="+_U);
        
        int numOfQueryInstantiations = (int)Math.pow(2,_numU);
        Hashtable h = new Hashtable(numOfQueryInstantiations );
        
        for (int p=0; p<numOfQueryInstantiations; p++) {

            String instRep = "";
            int[] binary = int2binary(p,_numU); //generate one instantiation
            Vector _copyOfF = (Vector)_Fremain.clone();
            Enumeration e = _U.elements();
            
            for (int r=0; r<_numU; r++) {
                String currentVariable = (String)e.nextElement();
                
                if (binary[r] == 0) {
                    _copyOfF = instantiateVariable(currentVariable,_FALSE,_copyOfF);
                    instRep = instRep.concat(currentVariable+_FALSE);
                }
                else {
                    _copyOfF = instantiateVariable(currentVariable,_TRUE,_copyOfF);
                    instRep = instRep.concat(currentVariable+_TRUE);
                    
                }
                if (r+1 < _numU) 
                    instRep = instRep.concat("&");
                
            }//for
            double answer = multiplyVector(_copyOfF);
            totalsum += answer;
            h.put(instRep,Double.toString(answer));
            
            if (DEBUG==1)System.out.println("P("+instRep+"&"+evidence+")="+answer);
        }//for

        Enumeration keys = h.keys();
        while(keys.hasMoreElements()) {
            String instRep = (String)keys.nextElement();
            String svalue = (String)h.get(instRep);
            double dvalue = Double.parseDouble(svalue);
            double finalanswer = dvalue/totalsum;
            System.out.println("P("+instRep+"|"+evidence+")="+finalanswer);
        }//while
    }//normaliseForEvidence
}//class
