import java.util.*;
import java.io.*;

public class Inference extends BeliefNet{

    int DEBUG;
    
    public Inference(Node[] ns,int doDebug) {
	super(ns);
        DEBUG=doDebug;
    }//constructor


    /********************** HELPER METHODS ************************/


    /***************************************************************/
    /**** Method initialiseFprime - returns a vector of all the ****/
    /**** elements in _F that mention _elim. Note that _F is    ****/
    /**** updated/modified in this method.                      ****/
    /***************************************************************/
    public Vector initialiseFprime(Vector _F, String _elim) {
	Vector _Fprime = new Vector();
	
	Enumeration e = _F.elements();
	while (e.hasMoreElements()) {
	    String exp = (String)e.nextElement();
	    if (mentions(exp,_elim)) {
		_Fprime.add(exp);
	    }
	}//while
	_F.removeAll(_Fprime); //extract means remove

	if (DEBUG==1) {
	    System.out.println("F="+_F);
	    System.out.println("F'="+_Fprime);
	}
	return _Fprime;
    }//initialiseFprime

    

    /*****************************************************************/
    /**** Method initialiseF - calls a method in class BeliefNet  ****/
    /**** which returns a vector of the string representations    ****/
    /**** for all the nodes/variables in the network. For example ****/
    /**** the string representation of variable wetgrass would be ****/
    /**** wetgrass|rain&cloudy.                                   ****/
    /*****************************************************************/
    public Vector initialiseF(String includeTheseInstantiations) {
        Vector f = getVectorRepresentation();
	if (includeTheseInstantiations.equals("")) return f;
        Hashtable _S = makeHashtableOfStrRep(includeTheseInstantiations);
        if (DEBUG==1)System.out.println("initialiseF: _S="+_S);
        Enumeration e = _S.keys();

	while (e.hasMoreElements()) {
	    String var = (String)e.nextElement();
	    String varValue = (String)_S.get(var);
	    f = instantiateVariable(var,varValue,f);
	}
	return f;
    }//initialiseF
    
    /*****************************************************************/
    /**** Method initialiseZ - returns a vector of all the nodes  ****/
    /**** or variables in the bayesNet except the ones contained  ****/
    /**** in minusThese                                           ****/
    /*****************************************************************/
    public Vector initialiseZ(String minusThese) {
	Vector Z = new Vector(_nn);
	for (int i=0; i<_nn; i++) {
	    if (!mentions(minusThese,_ns[i].getNodeName())) 
		Z.add(_ns[i].getNodeName());
	}//for
	return Z;
    }//initialiseF

    /******************************************************************/
    /**** Method - puts the string representation into a hashtable ****/
    /**** where the variables are mapped to their instantiations   ****/
    /******************************************************************/
    public Hashtable makeHashtableOfStrRep(String stringRep) {
	StringTokenizer st = new StringTokenizer(stringRep,"|&=");
	int num = st.countTokens()/2;
	Hashtable U = new Hashtable();

	for (int i=0; i<num; i++) {
	    U.put((Object)st.nextToken(),"="+(Object)st.nextToken());

	}
	return U;
    }//initialiseU

    /***********************************************************************************/
    /****  Method countUninstantiated - returns a vector of distinct uninstantiated ****/
    /****  variables/nodes in factors not including notThisVar                      ****/
    /***********************************************************************************/
    public Vector countUninstantiated(Vector factors,String notThisVar) {
	Enumeration e = factors.elements();
	Vector v = new Vector();
	while(e.hasMoreElements()) {
	    String key = (String)e.nextElement();
	    StringTokenizer st = new StringTokenizer(key,"|&");
	    int numTokens = st.countTokens();
	    for (int i=0; i<numTokens; i++) {
		String stringOfVector = v.toString();
		String token = st.nextToken();

		if ( !mentions(token,notThisVar) && !mentions(stringOfVector,token) )
		    if ( !mentions(token,_TRUE) && !mentions(token,_FALSE))
			v.add(token);
	    }//for
	}//while
	return v;
    }//countUninstantiated

    /*******************************************************************************/
    /****  Method instantitateVariable - for each factor in Vector factors;     ****/
    /****  the method assigns any instance of var that is found the value       ****/
    /****  varValue. For example, if rain&cloudy&sprinkler is one of the factors****/
    /****  and var="rain" and varValue=_FALSE then we change the factor to      ****/
    /****  rain=false&cloudy&sprinkler.  NOTE that varValue has to be one of the****/
    /****  2 predifined strings _TRUE or _FALSE                                 ****/
    /*******************************************************************************/
    public Vector instantiateVariable(String var, String varValue, Vector factors) {

	if (factors.isEmpty()) {
	    System.err.println("instantiateVariable: error, parameter factors is empty");
	    System.exit(-1);
	}

	Vector _r = new Vector();

	Enumeration _e = factors.elements();
	while (_e.hasMoreElements()) {
	    String _s = (String)_e.nextElement();
	    int end=0;
	    
	    //this assumes there is only one instance of a variable
	    //in every factor. 
	    if (mentions(_s,var)) {
		for (int i=0; i<_s.length(); i++)
		    if (_s.regionMatches(true,i,var,0,var.length()))
			end = i+var.length();
		
		String prefix = _s.substring(0,end);
		String suffix = _s.substring(end,_s.length());
	    
		_s = prefix+varValue+suffix;
	    }//if
	    _r.add((Object)_s);
	}//while
	return _r;
    }//instantiateVariable

    /********************************************************************************/
    /**** Method findInstantiation - finds the instantiation associated with var ****/
    /**** Assumes that all instances of the same variable are instantiated with  ****/
    /**** the same value. Returns the pre-defined strings _TRUE or _FALSE        ****/
    /**** The method is overloaded and can be passed either a vector or string.  ****/
    public String findInstantiation(Vector v, String var) {
	if (v.isEmpty()) {
	    System.err.println("findInstantiation: error, parameter vector is empty");
	    System.exit(-1);
	}

	Enumeration _e = v.elements();
	while (_e.hasMoreElements()) {
	    String _s = (String)_e.nextElement();
	    if (mentions(_s,var)) 
		return findInstantiation(_s, var);
	}
	System.err.println("findInstantion: "+var+" was not found in Vector="+v);
	System.exit(-1);
	return null;
    }//findInstantiation

    public String findInstantiation(String strRep, String var) {
	int endOfVarIndex = -1;

	for (int i=0; i<strRep.length(); i++)
	    if (strRep.regionMatches(i,var,0,var.length()))
		endOfVarIndex = i + var.length();

	if (strRep.regionMatches(endOfVarIndex,_TRUE,0,_TRUE.length()))
	    return _TRUE;
	if (strRep.regionMatches(endOfVarIndex,_FALSE,0,_FALSE.length()))
	    return _FALSE;
	System.err.println("findInstantiation: error, instantiation for "+var+" not found");
	System.exit(-1);
	return null;
    }//findInstantiation

    /**********************************************************************/
    /****  Method mentions - if var is a substring of exp return true  ****/
    /**********************************************************************/
    public boolean mentions(String exp, String var) {
	for (int i=0; i<exp.length(); i++) {
	    if (exp.regionMatches(i,var,0,var.length())) return true;
	}
	return false;
    }//mentions

    //returns true if v MENTIONS s, ie checks substrings in v
    public boolean vectorMentionsString(Vector v, String s){
        if (v.contains(s)) return true;
        for (int i=0; i<v.size(); i++) {
            String check = (String)v.get(i);
            if (mentions(check,s)) return true;
        }
        return false;
    }//vectorContains
    
    /**********************************************************************/
    /****  Method contains - use when the order is important. For      ****/
    /****  example if you want to check if v  contains the string s:   ****/
    /****  rain&sprinkler and the hashtable contains sprinkler&rain,   ****/
    /****  the contains method will return true even though the order  ****/
    /****  is reversed. Assuming the method is passed uninstantiated   ****/
    /****  variables in s.                                             ****/
    /**********************************************************************/
    public boolean contains(Vector v, String s) {
	if (v.contains(s)) return true;

	StringTokenizer st = new StringTokenizer(s,"|");
	if (st.countTokens() == 2) {
	    String query = st.nextToken();
	    String evidence = st.nextToken();
	    String[] queryArray = Permutations(query);
	    String[] evidenceArray = Permutations(evidence);
	    
	    for (int i=0; i<queryArray.length; i++)
		for (int j=0; j<evidenceArray.length; j++) {
		    String factor = queryArray[i];
		    factor = factor.concat("|"+evidenceArray[j]);
		    if (v.contains(factor)) return true;
		}//for
	}//if
	else {
	    String query = st.nextToken();
	    String[] queryArray = Permutations(query);
	    for (int i=0; i<queryArray.length; i++)
		if (v.contains(queryArray[i]))return true;
	}//else
	return false;
    }//contains
    
    /***********************************************************************/
    /**** uninstantiates all instances of var in vector v and returns it ***/
    /***********************************************************************/
    public Vector unInstantiate(Vector v, String var) {
	Vector r = new Vector();
	Enumeration e = v.elements();
	
	while (e.hasMoreElements()) {
	    String s = String.valueOf(e.nextElement());
	    String uninstantiated = "";
	    int beginOfInst=0;
	    boolean match = false;
	    for (int i=0; i<s.length(); i++) {
		if (s.regionMatches(true,i,var,0,var.length())) {
		    match = true;
		    beginOfInst=i+var.length();
		    if (s.regionMatches(true,beginOfInst,_TRUE,0,_TRUE.length())) {
			String prefix = s.substring(0,beginOfInst);
			String suffix = s.substring(beginOfInst+_TRUE.length(),s.length());
			if(!r.contains(prefix+suffix))r.add(prefix+suffix);
			break;
		    }//if
		    else if (s.regionMatches(true,beginOfInst,_FALSE,0,_FALSE.length())) {
			String prefix = s.substring(0,beginOfInst);
			String suffix = s.substring(beginOfInst+_FALSE.length(),s.length());
			if(!r.contains(prefix+suffix))r.add(prefix+suffix);
			break;
		    }//else
		}//if
	    }//for
	    if (!match && !r.contains(s)) r.add(s);
	}//while
	if (DEBUG==1) System.out.println("uninstantiated "+var+" to give r="+r);
	return r;
    }//unInstantiate





    /************************************* UNUSED METHODS **********/


    /*****************************************************************/
    /****  Method unInstantiate - removes from every entry in ht  ****/
    /****  substrings of the form "=true" and "=false", returning ****/
    /****  a Vector of the distinct substrings that are formed    ****/
    /****  ie, the entry "cloudy=true&rain=false" would be        ****/
    /****  reduced to "cloudy&rain"                               ****/
    /*****************************************************************/
    public Vector unInstantiate(Hashtable ht) {
	Vector r = new Vector();
	Enumeration e = ht.keys();

	while (e.hasMoreElements()) {
	    String s = String.valueOf(e.nextElement());
	    String uninstantiated = "";
	    int begin=0;

	    for (int i=0; i<s.length(); i++) {

		if (s.regionMatches(true,i,_TRUE,0,_TRUE.length())) {
		    uninstantiated = uninstantiated.concat(s.substring(begin,i));
		    begin = i+_TRUE.length();
		}//if
		else if (s.regionMatches(true,i,_FALSE,0,_FALSE.length())) {
		    uninstantiated = uninstantiated.concat(s.substring(begin,i));
		    begin = i+_FALSE.length();
		}//else
	    }//for
	    if (!r.contains((Object)uninstantiated))
		r.add((Object)uninstantiated);
	}//while
	return r;
    }//unInstantiate



}//class
