import java.util.*;
import java.io.*;

public class Menu {
    final String _TRUE = "=true";
    final String _FALSE = "=false";

    BeliefNet _bn;
    Node[] _ns;
    int _numNodes;

    //Constructor for Class Menu
   public Menu(int debug) throws IOException {
	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader bf = new BufferedReader(isr);

	System.out.println("Create a NEW Belief-Net? (1)");
        System.out.print("LOAD a Belief-Net from disk? (2) ");

        if ((bf.readLine()).equals("2")) {

            System.out.print("Enter the filename of the Belief-Net? ");
	    try {
		ObjectReader or = ObjectReader.openFileForReading(bf.readLine());
		_bn = (BeliefNet) or.readObject();
		_ns = _bn.getNodeStructure();
		_numNodes = _ns.length;
		or.close();
	    }
	    catch(Exception e) {
		System.out.println("\nUnable to open/read from the specified file.");
		System.out.println("This could be because changes have been made to either the Node or BeliefNet classes.");
		System.exit(-1);
	    }
            System.out.print("Do you want to display the Belief-Net? (y/n) ");
            String answer = bf.readLine();
            if (answer.equals("yes")||answer.equals("y"))
                for (int i=0; i<_numNodes; i++) _ns[i].displayNode();
        }//if
        else {
            _bn = new BeliefNet();
	         _ns = _bn.getNodeStructure();
	         _numNodes = _ns.length;

            System.out.print("Would you like to save the Belief-Net to disk? (y/n) ");
            String answer = bf.readLine();
            if (answer.equals("yes")||answer.equals("y")) {
                System.out.print("Enter a filename?");

                ObjectWriter ow = ObjectWriter.openFileForWriting(bf.readLine());
                ow.writeObject((Object)_bn);
                ow.close();

            }
        }//else
	
	while (true) {
	    System.out.println("\nMake a selection:");
	    System.out.println("(1) Exact Inference");
	    System.out.println("(2) Approximate Inference");
	    System.out.println("(3) Quit");
	    
	    String choice = bf.readLine();
	    if (choice.equals("1"))
		ExactInference(debug);
	    if (choice.equals("2"))
		ApproxInference(debug);
	    if (choice.equals("3"))
		break;
	}

   }//Constructor

    public void ApproxInference(int debug) throws IOException{
	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader bf = new BufferedReader(isr);

	String query = "";

	while (true) {
	    System.out.println("Choose a query variable.");
	    for (int i=0; i<_numNodes; i++)
		System.out.println("("+(i+1)+") "+_ns[i].getNodeName());
	    System.out.println("("+(_numNodes+1)+") No more query variables.");
	    
	    int choice = Integer.parseInt(bf.readLine()) - 1;
	    if (choice == _numNodes) break;

	    System.out.println(_ns[choice].getNodeName()+"= true or false? (t/f)");
	    String inst = bf.readLine();

	    if (!mentions(query,_ns[choice].getNodeName())) {
		if (!query.equals(""))
		    query = query.concat("&");

		query = query.concat(_ns[choice].getNodeName());

		if (inst.equalsIgnoreCase("t") || inst.equalsIgnoreCase("true"))
		    query = query.concat(_TRUE);
		else 
		    query = query.concat(_FALSE);

	    }

	}//while

        if (query.equals("")) {
            System.out.println("You have not selected any query variables - exiting...");
            System.exit(-1);
        }
        
	String evidence = "";
	while (true) {
	    System.out.println("Choose a variable to provide as evidence.");
	    for (int i=0; i<_numNodes; i++)
		System.out.println("("+(i+1)+") "+_ns[i].getNodeName());
	    System.out.println("("+(_numNodes+1)+") No more evidence variables");
	    
	    int choice = Integer.parseInt(bf.readLine()) - 1;
	    if (choice == _numNodes) break;
	    
	    if (!mentions(evidence,_ns[choice].getNodeName())) {
		if (!evidence.equals(""))
		    evidence = evidence.concat("&");
		evidence = evidence.concat(_ns[choice].getNodeName());
		System.out.println(_ns[choice].getNodeName()+"= true or false? (t/f)");
		String inst = bf.readLine();
		if (inst.equalsIgnoreCase("t") || inst.equalsIgnoreCase("true"))
		    evidence = evidence.concat(_TRUE);
		else 
		    evidence = evidence.concat(_FALSE);
	    }//if
	}//while
	System.out.println("Choose the number of iterations:");
	int numIter = Integer.parseInt(bf.readLine());

	System.out.println("Rejection Sampling (1)");
	System.out.println("Likelihood Weighting (2)");
        System.out.println("Gibbs Sampling (3)");
	String choice = bf.readLine();
	ApproxInference _ap = new ApproxInference(_ns,debug);

	if (choice.equals("1"))
	    _ap.rejectionSampling(query,evidence,numIter);
	else if (choice.equals("2"))
	    _ap.likelihoodWeighting(query,evidence,numIter);
    else
        _ap.gibbs(query,evidence,numIter);
        

	//ai.rejectionSampling("cloudy=false","wetgrass=true&sprinkler=false");
    }//approx inference

    
    public void ExactInference(int debug) throws IOException {
	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader bf = new BufferedReader(isr);

	String query = "";

	while (true) {
	    System.out.println("Choose a query variable.");
	    for (int i=0; i<_numNodes; i++)
		System.out.println("("+(i+1)+") "+_ns[i].getNodeName());
	    System.out.println("("+(_numNodes+1)+") No more query variables.");
	    
	    int choice = Integer.parseInt(bf.readLine()) - 1;
	    if (choice == _numNodes) break;
            
	    if (!mentions(query,_ns[choice].getNodeName())) {
		if (!query.equals(""))
		    query = query.concat("&");
		query = query.concat(_ns[choice].getNodeName());
	    }

	}

	if (query.equals("")) {
	    System.err.println("You have not selected any query variables - exiting...");
	    System.exit(-1);
	}
        
	ExactInference _ei = new ExactInference(_ns,debug);

	System.out.println("Do you want to observe a variable? ie, provide evidence? (y/n)");
	String ans = bf.readLine();

	if (ans.equalsIgnoreCase("Y") || ans.equalsIgnoreCase("yes")) {
	    String evidence = "";

	    while (true) {
		System.out.println("Choose a variable to provide as evidence.");
		for (int i=0; i<_numNodes; i++)
		    System.out.println("("+(i+1)+") "+_ns[i].getNodeName());
		System.out.println("("+(_numNodes+1)+") No more evidence variables");

		int choice = Integer.parseInt(bf.readLine()) - 1;
		if (choice == _numNodes) break;

		if (!mentions(evidence,_ns[choice].getNodeName())) {
		    if (!evidence.equals(""))
			evidence = evidence.concat("&");
		    evidence = evidence.concat(_ns[choice].getNodeName());
		    System.out.println(_ns[choice].getNodeName()+"= true or false? (t/f)");
		    ans = bf.readLine();
		    if (ans.equalsIgnoreCase("t") || ans.equalsIgnoreCase("true"))
			evidence = evidence.concat(_TRUE);
		    else 
			evidence = evidence.concat(_FALSE);
		}//if
	    }//while
	    _ei.variableEliminationEvidence(query,evidence);
	}//variable elimination with evidence
	else {
	    _ei.variableElimination(query);
	}//variable elimination with NO evidence

    }//ExactInference

    public boolean mentions(String exp, String var) {
	for (int i=0; i<exp.length(); i++) {
	    if (exp.regionMatches(i,var,0,var.length())) return true;
	}
	return false;
    }

	 //*IGNORE* this constructor
    public Menu(boolean hello) {
	ObjectReader or = ObjectReader.openFileForReading("network");
	_bn = (BeliefNet) or.readObject();
	_ns = _bn.getNodeStructure();
	_numNodes = _ns.length;
	ApproxInference _ap = new ApproxInference(_ns,1);
	_ap.rejectionSampling("cloudy=false","wetgrass=true&sprinkler=false",1000);
    }

    /**************************************************************************/
    /****                                 MAIN                             ****/
    /**************************************************************************/
    public static void main(String[] args) throws IOException{
        int debug = 0;
        if (args.length > 0)
            if (args[0].equalsIgnoreCase("-d")) debug = 1;
            
	     Menu m = new Menu(debug);
    }

}//Class Menu
