import java.util.*;
import java.io.*;
public class Node implements Serializable {
    
    private String variable_name;    //identifier for this object
    private String[] parents_ref;    //references to parent nodes using their variable_name
    private String[] children_ref;   //references to child nodes using their variable_name
    private int num_parents;         //a count of the number of parents
    private int num_children;        //a count of the number of children
    private Hashtable CPT;           //Conditional Probability Table for this node
    
    //Constructor for Node Object
    public Node (String vn) {

	    variable_name = vn;

	try {
	    addParent();
	    addChild();
	    CPT = new Hashtable();
	    addProbabilityTable();
	}
	catch (IOException ioe) {
	    ioe.printStackTrace();
	}
    }//Constructor

    public void addParent() throws IOException
    {

	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader bf = new BufferedReader(isr);
	
	System.out.print("How many parents does this variable have? ");
	num_parents = Integer.parseInt(bf.readLine());
	parents_ref = new String[num_parents];
	
	for (int i=0; i<num_parents; i++) {
	    System.out.print("Enter parent name "+(i+1)+" ? ");
	    parents_ref[i] = bf.readLine();
	}
    }//addParent

    public void addChild() throws IOException
    {
	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader bf = new BufferedReader(isr);
	
	System.out.print("How many children does this variable have? ");
	num_children = Integer.parseInt(bf.readLine());
	children_ref = new String[num_children];
	
	for (int i=0; i<num_children; i++) {
	    System.out.print("Enter child name "+(i+1)+" ? ");
	    children_ref[i] = bf.readLine();
	}
    }//addChild

    //Method to calculate the size of the CPT for this Node
    //Must be called after adding ALL children and parent references
    public int calcProbTableSize() 
    {
        if (parents_ref == null) return 1;
        return (int)Math.pow(2,num_parents);
    }//calcProbTableSize
        
    public void addProbabilityTable() throws IOException
    {
	InputStreamReader isr = new InputStreamReader(System.in);
	BufferedReader bf = new BufferedReader(isr);

        System.out.println("Please enter the following probabilities - \n");
        
        for (int i=0; i<calcProbTableSize(); i++) {

            System.out.print ("P(");

            String CP_true = variable_name+"=true";
	    String CP_false = variable_name+"=false";
	    String evidence = "";

	    if (num_parents > 0) CP_true = CP_true.concat("|");
	    if (num_parents > 0) CP_false = CP_false.concat("|");

            int[] binary = int2binary(i);

            for (int j=0; j<num_parents; j++) {
                if(binary[j]==0)evidence=evidence.concat(parents_ref[j]+"=false");
                else evidence = evidence.concat (parents_ref[j]+"=true");
                
                if (j+1 < num_parents) evidence = evidence.concat("&");
            }

	    CP_true = CP_true.concat(evidence);
	    CP_false = CP_false.concat(evidence);

            System.out.print (CP_true+") = ? ");

	    String true_response = bf.readLine();
	    double true_double = Double.parseDouble(true_response);
	    double false_double = 1 - true_double;
	    String false_response = String.valueOf(false_double);

            //add the entry to the probability table
            CPT.put(CP_true,true_response);
	    CPT.put(CP_false,false_response);

        }//for
        
    }//addProbabilityTable

    //Helper method for function addProbabilityTable
    //sets the length of the binary string to be the number of parents
    public int[] int2binary (int value) 
    {
        String binary ="";
        int[] rem = new int[num_parents];

	for (int h=0; h<num_parents; h++) rem[h] = 0;

	for (int i=num_parents-1; value > 0; i--) {

	    rem[i] = value%2;
	    value = value/2;
	}        

	return rem;
    }//int2binary
    
    
    //I don't think this method is being used
    //Returns probabilities specific to this Node
    public double getProbability(String stmt) {
	if (CPT.containsKey(stmt)) {
	    return Double.parseDouble((String)CPT.get((Object)stmt));
	}//if

	StringTokenizer st = new StringTokenizer (stmt,"|");

	if (st.countTokens() == 1) {
	    System.err.println("Bad parameter at getProbability, try another Node object!");
	    System.exit(-1);
	}//if
	else if (st.countTokens() == 2) {
	    String query = st.nextToken();
	    String evidence = st.nextToken();
	    System.out.println("QUERY -"+query+"-");
	    System.out.println("EVIDENCE -"+evidence+"-");

	    StringTokenizer st_query = new StringTokenizer(query);
	    
	    //this is not true if 2 or more query vars are possible in query stmt.
	    if (!variable_name.equals(st_query.nextToken("="))) {
		System.out.println("getProbability() method called on wrong Node object");
		System.exit(-1);
	    }//if
	    
	}//else if

	return -1.0;
    }


    public String[] getChildren() {
	return children_ref;
    }

    public String[] getParents() {
	return parents_ref;
    }

    public Hashtable getCPT() {
	return CPT;
    }

    public String getNodeName() {
	return variable_name;
    }

    public void displayNode() {
	int i;
	int length = 80;
	String temp = "";

	System.out.println();
	for (i=0; i<length; i++) System.out.print("-");

	temp = "\n"+"|  Node: "+variable_name;
	System.out.print(temp);
	for (i=0; i<length-temp.length();i++) System.out.print(" ");
	System.out.println("|");

	if (CPT.isEmpty()) {
	temp = "|  Empty Probability Table!";
	System.out.print(temp);
	for (i=0; i<length-temp.length();i++) System.out.print(" ");
	System.out.println("|");
	}
	else {
	    for (i=0; i<length; i++) System.out.print("-");
	    System.out.println();

	    temp = "|  Parents - ";
	    for (i=0; i<parents_ref.length; i++) 
		temp = temp.concat(parents_ref[i]+"  ");
	    System.out.print(temp);
	    for (i=0; i<length-temp.length()-1;i++) System.out.print(" ");
	    System.out.println("|");

	    temp = "|  Children - ";
	    for (i=0; i<children_ref.length; i++) 
		temp = temp.concat(children_ref[i]+"  ");
	    System.out.print(temp);
	    for (i=0; i<length-temp.length()-1;i++) System.out.print(" ");
	    System.out.println("|");

	    Enumeration e = CPT.keys();
	    while (e.hasMoreElements()) {
		Object key = e.nextElement();
		String prob = (String)CPT.get(key);
		temp = "|  P("+(String)key+") = "+prob;
		System.out.print(temp);
		for (i=0; i<length-temp.length()-1;i++) System.out.print(" ");
		System.out.println("|");
	    }//while
	    for (i=0; i<length; i++) System.out.print("-");
	}//else
	System.out.println();
    }
    
}//Node class
